import typing as t
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum

from starlette.requests import HTTPConnection, Request
from starlette.responses import Response

from .mixins import AnonymousUser
from .utils import decode_cookie, encode_cookie


class ProtectionLevel(Enum):
    Basic = 1
    Strong = 2


@dataclass
class Config:
    SESSION_NAME_FRESH = '_fresh'
    SESSION_NAME_ID = '_id'
    SESSION_NAME_KEY = '_user_id'
    SESSION_NAME_NEXT = 'next'

    REMEMBER_COOKIE_NAME = '_remember'
    REMEMBER_SECONDS_NAME = '_remember_seconds'
    EXEMPT_METHODS = {'OPTIONS'}

    protection_level: t.Optional[ProtectionLevel] = ProtectionLevel.Basic

    # Cookie configuration
    COOKIE_NAME = 'remember_token'
    COOKIE_DOMAIN: t.Optional[str] = None
    COOKIE_PATH: str = '/'
    COOKIE_SECURE: bool = False
    COOKIE_HTTPONLY: bool = True
    COOKIE_SAMESITE = None
    COOKIE_DURATION: timedelta = timedelta(days=365)

    @property
    def session_keys(self):
        return (
            self.SESSION_NAME_FRESH,
            self.SESSION_NAME_ID,
            self.SESSION_NAME_KEY,
            self.SESSION_NAME_NEXT,
            self.REMEMBER_COOKIE_NAME,
            self.REMEMBER_SECONDS_NAME,
        )


class LoginManager:
    _user_loader: t.Callable = None

    def __init__(
        self, redirect_to: str, secret_key: str, config: Config = None
    ):
        self.config = config or Config()
        self.anonymous_user_cls = AnonymousUser
        # Name of redirect view when user need to log in.
        self.redirect_to = redirect_to
        self.secret_key = secret_key

    def set_user_loader(self, callback: t.Callable):
        self._user_loader = callback

    @property
    def user_loader(self):
        assert self._user_loader is not None, \
            '`user_loader` is required'
        return self._user_loader

    def build_redirect_url(self, request: HTTPConnection):
        if '/' in self.redirect_to:
            return self.redirect_to
        return request.url_for(self.redirect_to)

    def protection_is_strong(self):
        return self.config.protection_level == ProtectionLevel.Strong

    def set_cookie(self, response: Response, user_id: t.Any):
        # if not isinstance(user_id, str):
        #     user_id = str(user_id)
        response.set_cookie(
            key=self.config.COOKIE_NAME,
            value=encode_cookie(user_id, self.secret_key),
            expires=int(self.config.COOKIE_DURATION.total_seconds()),
            path=self.config.COOKIE_PATH,
            domain=self.config.COOKIE_DOMAIN,
            secure=self.config.COOKIE_SECURE,
            httponly=self.config.COOKIE_HTTPONLY,
            samesite=self.config.COOKIE_SAMESITE
        )

    def get_cookie(self, cookie: str):
        return decode_cookie(cookie, self.secret_key)



import typing as t

from starlette.middleware.base import (
    BaseHTTPMiddleware, RequestResponseEndpoint
)
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from .backends import BaseAuthenticationBackend
from .login_manager import LoginManager


class AuthenticationMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        backend: BaseAuthenticationBackend,
        login_manager: LoginManager,
        secret_key: str,
        login_route: str = None,
        excluded_dirs: t.List[str] = None
    ):
        super().__init__(app)
        self.app = app
        self.backend = backend
        self.login_route = login_route
        self.secret_key = secret_key
        self.login_manager = login_manager
        self.excluded_dirs = excluded_dirs or []

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        scope = request.scope
        # Excluded prefix path. E.g. /static
        for prefix_dir in self.excluded_dirs:
            if scope['path'].startswith(prefix_dir):
                response = await call_next(request)
                return response

        user = await self.backend.authenticate(request)
        if not user or user.is_authenticated is False:
            request.scope['user'] = self.login_manager.anonymous_user_cls()
        else:
            request.scope['user'] = user

        response = await call_next(request)

        if user and user.is_authenticated:
            operation = request.session.get(
                self.login_manager.config.REMEMBER_COOKIE_NAME
            )
            if operation == 'set':
                self.login_manager.set_cookie(response, user.identity)

        return response



import asyncio
import functools
import inspect
import typing

from starlette.requests import Request
from starlette.responses import Response, RedirectResponse

from .utils import create_identifier, make_next_url

LOGIN_MANAGER_ERROR = 'LoginManager is not set'


def login_required(func: typing.Callable) -> typing.Callable:
    sig = inspect.signature(func)
    for idx, parameter in enumerate(sig.parameters.values()):
        if parameter.name == "request":
            break
    else:
        raise Exception(
            f'No "request" argument on function "{func}"'
        )   # pragma: no cover

    if asyncio.iscoroutinefunction(func):
        @functools.wraps(func)
        async def async_wrapper(
            *args: typing.Any, **kwargs: typing.Any
        ) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)

            login_manager = getattr(request.app.state, 'login_manager', None)
            assert login_manager is not None, LOGIN_MANAGER_ERROR

            if request.method in login_manager.config.EXEMPT_METHODS:
                return await func(*args, **kwargs)    # pragma: no cover

            user = request.scope.get('user')
            if not user or getattr(user, 'is_authenticated', False) is False:
                redirect_url = make_next_url(
                    login_manager.build_redirect_url(request),
                    str(request.url)
                )
                return RedirectResponse(redirect_url, status_code=302)
            else:
                return await func(*args, **kwargs)
        return async_wrapper
    else:
        @functools.wraps(func)
        def sync_wrapper(*args: typing.Any, **kwargs: typing.Any) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)

            login_manager = getattr(request.app.state, 'login_manager', None)
            assert login_manager is not None, LOGIN_MANAGER_ERROR

            if request.method in login_manager.config.EXEMPT_METHODS:
                return func(*args, **kwargs)    # pragma: no cover

            user = request.scope.get('user')
            if not user or getattr(user, 'is_authenticated', False) is False:
                redirect_url = make_next_url(
                    login_manager.build_redirect_url(request),
                    str(request.url)
                )
                return RedirectResponse(redirect_url, status_code=302)
            else:
                return func(*args, **kwargs)
        return sync_wrapper


def fresh_login_required(func: typing.Callable) -> typing.Callable:
    sig = inspect.signature(func)
    for idx, parameter in enumerate(sig.parameters.values()):
        if parameter.name == "request":
            break
    else:
        raise Exception(
            f'No "request" argument on function "{func}"'
        )   # pragma: no cover

    if asyncio.iscoroutinefunction(func):
        @functools.wraps(func)
        async def async_wrapper(
            *args: typing.Any, **kwargs: typing.Any
        ) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)
            login_manager = getattr(request.app.state, 'login_manager', None)
            assert login_manager is not None, 'LoginManager state is not set'

            if request.method in login_manager.config.EXEMPT_METHODS:
                return await func(*args, **kwargs)    # pragma: no cover

            session_fresh = login_manager.config.SESSION_NAME_FRESH

            user = request.scope.get('user')
            if not user \
                    or getattr(user, 'is_authenticated', False) is False \
                    or request.session.get(session_fresh, False) is False:
                request.session[
                    login_manager.config.SESSION_NAME_ID
                ] = create_identifier(request)

                return RedirectResponse(make_next_url(
                    login_manager.build_redirect_url(request),
                    str(request.url)
                ), status_code=302)
            else:
                return await func(*args, **kwargs)
        return async_wrapper
    else:
        @functools.wraps(func)
        def sync_wrapper(*args: typing.Any, **kwargs: typing.Any) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)
            login_manager = getattr(request.app.state, 'login_manager', None)
            assert login_manager is not None, 'LoginManager state is not set'

            if request.method in login_manager.config.EXEMPT_METHODS:
                return func(*args, **kwargs)    # pragma: no cover

            session_fresh = login_manager.config.SESSION_NAME_FRESH

            user = request.scope.get('user')
            if not user \
                    or getattr(user, 'is_authenticated', False) is False \
                    or request.session.get(session_fresh, False) is False:
                request.session[
                    login_manager.config.SESSION_NAME_ID
                ] = create_identifier(request)

                return RedirectResponse(make_next_url(
                    login_manager.build_redirect_url(request),
                    str(request.url)
                ), status_code=302)
            else:
                return func(*args, **kwargs)
        return sync_wrapper



import hmac
import typing
from datetime import timedelta
from hashlib import sha512
from urllib.parse import quote, urlparse, urlunparse

from starlette.requests import Request

from .mixins import UserMixin, AnonymousUser

LOGIN_MANAGER_ERROR = 'LoginManager is not set'


async def login_user(
    request: Request, user: UserMixin, remember: bool = False,
    duration: timedelta = None, fresh: bool = True
) -> bool:
    assert request.scope.get('app') is not None, \
        'Invalid Starlette app'
    login_manager = getattr(request.app.state, 'login_manager', None)
    assert login_manager is not None, LOGIN_MANAGER_ERROR

    login_manager = getattr(request.app.state, 'login_manager', None)
    assert login_manager is not None, LOGIN_MANAGER_ERROR
    assert user.identity is not None, \
        'user identity implementation is required'

    request.session[
        login_manager.config.SESSION_NAME_KEY
    ] = user.identity
    request.session[
        login_manager.config.SESSION_NAME_FRESH
    ] = fresh
    request.session[
        login_manager.config.SESSION_NAME_ID
    ] = create_identifier(request)
    if remember:
        request.session[
            login_manager.config.REMEMBER_COOKIE_NAME
        ] = 'set'
        if duration is not None:
            request.session[
                login_manager.config.REMEMBER_SECONDS_NAME
            ] = duration.total_seconds()
    request.scope['user'] = user
    return True


async def logout_user(request: Request) -> None:
    assert request.scope.get('app') is not None, \
        'Invalid Starlette app'
    login_manager = getattr(request.app.state, 'login_manager', None)
    assert login_manager is not None, LOGIN_MANAGER_ERROR

    session_key = login_manager.config.SESSION_NAME_KEY
    session_fresh = login_manager.config.SESSION_NAME_FRESH
    session_id = login_manager.config.SESSION_NAME_ID
    remember_cookie = login_manager.config.REMEMBER_COOKIE_NAME

    if session_key in request.session:
        request.session.pop(session_key)

    if session_fresh in request.session:
        request.session.pop(session_fresh)

    if session_id in request.session:
        request.session.pop(session_id)

    if remember_cookie in request.cookies:
        request.session[remember_cookie] = 'clear'
        remember_seconds = login_manager.config.REMEMBER_SECONDS_NAME
        if remember_seconds in request.session:
            request.session.pop(remember_seconds)

    request.scope['user'] = AnonymousUser()


def encode_cookie(payload: str, key: str) -> str:
    return f'{payload}|{_cookie_digest(payload, key=key)}'


def decode_cookie(cookie: str, key: str) -> typing.Optional[str]:
    try:
        payload, digest = cookie.rsplit('|', 1)
        if hasattr(digest, 'decode'):
            digest = digest.decode('ascii')
    except ValueError:
        return

    if hmac.compare_digest(_cookie_digest(payload, key=key), digest):
        return payload


def make_next_url(
    redirect_url: str, next_url: str = None
) -> typing.Optional[str]:
    if next_url is None:
        return redirect_url

    r_url = urlparse(redirect_url)
    n_url = urlparse(next_url)

    if (not r_url.scheme or r_url.scheme == n_url.scheme) and \
            (not n_url.netloc or n_url.netloc == n_url.netloc):
        param_next = urlunparse((
            '', '', n_url.path, n_url.params, n_url.query, ''
        ))
    else:
        param_next = next_url

    if param_next:
        param_next = '='.join(('next', quote(param_next)))

    if r_url.query:
        result_url = r_url._replace(query='&'.join((r_url.query, param_next)))
    else:
        result_url = r_url._replace(query=param_next)
    return urlunparse(result_url)


def _get_remote_address(request: Request) -> typing.Optional[str]:
    address = request.headers.get('X-Forwarded-For')
    if address is not None:
        address = address.split(',')[0].strip()
    else:
        client = request.scope.get('client')
        if client is not None:
            address = client[0]
        else:
            address = None
    return address


def create_identifier(request) -> str:
    user_agent = request.headers.get('User-Agent')
    if user_agent is not None:
        user_agent = user_agent.encode('utf-8')
    base = f'{_get_remote_address(request)}|{user_agent}'
    h = sha512()
    h.update(base.encode('utf8'))
    return h.hexdigest()


def _secret_key(secret_key: typing.Union[bytes, str]) -> bytes:
    """ensure bytes"""
    if isinstance(secret_key, str):
        return secret_key.encode('latin1')
    return secret_key


def _cookie_digest(payload: str, key: str) -> str:
    key = _secret_key(key)
    if not isinstance(payload, str):
        payload = str(payload)
    return hmac.new(key, payload.encode('utf-8'), sha512).hexdigest()



import asyncio
import typing as t

from starlette.requests import HTTPConnection

from .login_manager import LoginManager, ProtectionLevel
from .mixins import UserMixin


class BaseAuthenticationBackend:
    # name: str = None

    async def authenticate(self, request: HTTPConnection):
        ...     # pragma: no cover


class SessionAuthBackend(BaseAuthenticationBackend):
    def __init__(self, login_manager: LoginManager):
        self.login_manager = login_manager

    async def authenticate(
        self, request: HTTPConnection
    ) -> t.Optional[UserMixin]:
        # Load user from session
        session_key = self.login_manager.config.SESSION_NAME_KEY
        user_id = request.session.get(session_key)

        # Using Strong protection
        if self.login_manager.protection_is_strong():
            for key in self.login_manager.config.session_keys:
                try:
                    request.session.pop(key)
                except KeyError:
                    pass
            request.session[
                self.login_manager.config.REMEMBER_COOKIE_NAME
            ] = 'clear'
        else:
            request.session[
                self.login_manager.config.SESSION_NAME_FRESH
            ] = False

        if user_id is None and request.session.get(
            self.login_manager.config.REMEMBER_COOKIE_NAME, 'clear'
        ) != 'clear':
            cookie = request.cookies.get(
                self.login_manager.config.COOKIE_NAME
            )
            if cookie:
                user_id = self.login_manager.get_cookie(cookie)
                user_id = int(user_id)
                request.session[
                    self.login_manager.config.SESSION_NAME_FRESH
                ] = False

        if user_id is None:
            return
        elif asyncio.iscoroutinefunction(self.login_manager.user_loader):
            user = await self.login_manager.user_loader(request, user_id)
        else:
            user = self.login_manager.user_loader(request, user_id)

        if user is not None:
            return user
